"""Miscellaneous functions for the client.
"""
import numpy as np
import pandas as pd
import itertools
import inspect
import types
import random
import warnings
from typing import List, Any, Dict, Union
from functools import wraps, partial

class UtilsMixin:
    """Various utilties
    """
    def generate_vector(self, vector_length: int, num_of_constant_values: int=0):
        """
        Generate a random vector based on length

        Args:
            vector_length:
                Length of a vector.
            num_of_constant_values
                The number of constant values in a vector. This is to help make the
                vectors more similar

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> vi_client.generate_vector(vector_length=20)
        """
        return np.random.rand(vector_length - num_of_constant_values).tolist() + [0.5] * num_of_constant_values

    @staticmethod
    def results_to_df(data):
        """
        Used for converting json responses of search results, etc
        Args:
            data:
                data returned from search, retrieve_documents, etc
        """
        if "results" in data:
            return pd.DataFrame(data["results"])
        elif "documents" in data:
            return pd.DataFrame(data["documents"])
        elif type(data) == dict:
            return pd.DataFrame([data])
        elif type(data) == list:
            return pd.DataFrame(data)
        else:
            print("missing field")

    def clean_results(self, results):
        if 'results' in results:
            return results['results']
        elif 'documents' in results:
            return results['documents']
        return results

    def results_pretty(self, results: Union[Dict, List], field: str):
        """
        Make results pretty in a Pandas Dataframe.
        Args:
            results:
                JSON of results
            field:
                Field to prettify results.

        Example:
            >>> result = vi_client.search(collection_name, text_encoder.encode('No more tissue paper'),
                field='OriginalTweet_vector_', page_size=5)
            >>> vi_client.results_pretty(result, 'OriginalTweet')
        """
        pd.set_option('display.max_colwidth', 10000)
        if 'results' in results.keys():
            results = results['results']
        return pd.DataFrame({self.get_field(field, x) for x in results}, columns=[field])

    @staticmethod
    def flatten_list(list_of_lists: List[List]):
        """
        Flatten a nested list

        Args:
            list_of_lists:
                A list of lists to flatten to make into 1 list.

        Example:
            >>> from vectorai.client import ViClient
            >>> ViClient.flatten_list([[0, 1], [2, 3]])
        """
        return list(itertools.chain.from_iterable(list_of_lists))

    @staticmethod
    def rename(object: Any, value: str):
        """
            Rename an object.
            Args:
                object:
                    Any Python instance/variable
                value:
                    Name of new instance

            Example:
                >>> from vectorai.client import ViClient
                >>> ViClient.rename(model, 'new_name')
        """
        object.__name__ = value

    @staticmethod
    def set_name(object: Any, value: str):
        """
            Rename an object.
            Args:
                object:
                    Any Python object

            Example:
                >>> from vectorai.client import ViClient
                >>> ViClient.set_name(model, 'new_name')
        """
        object.__name__ = value

    @staticmethod
    def get_name(object):
        """
            Get the name attribute of an object.
            Args:
                object:
                    Any Python object
            Returns:
                object name:
                    Name of object

            Example:
                >>> get_name(try_it)
        """
        return object.__name__

    @staticmethod
    def is_in_ipython():
        """
        Determines if current code is executed within an ipython session.
        """
        is_in_ipython = False
        # Check if the runtime is within an interactive environment, i.e., ipython.
        try:
            from IPython import get_ipython  # pylint: disable=import-error
            if get_ipython():
                is_in_ipython = True
        except ImportError:
            pass  # If dependencies are not available, then not interactive for sure.
        return is_in_ipython

    def is_in_notebook(self) -> bool:
        """
        Determines if current code is executed from an ipython notebook.
        """
        is_in_notebook = False
        if self.is_in_ipython():
            # The import and usage must be valid under the execution path.
            from IPython import get_ipython
            if 'IPKernelApp' in get_ipython().config:
                is_in_notebook = True
        return is_in_notebook

    def progress_bar(self, documents: List, total: int=None, show_progress_bar=True):
        """
        Returns a progress bar. depending on if notebook is available or not.

        Args:
            documents:
                A list of documents (Python dictionaries)

            Total:
                Length of documents/size of update

        Returns:
            Tqdm Progress Bar

        Example:
            for i in UtilsMixin.progress_bar(documents):
                ...
        """
        if not show_progress_bar:
            return documents

        if total is None:
            total = len(documents)

        if self.is_in_notebook():
            from tqdm.notebook import tqdm
            return tqdm(documents, total=total)
        else:
            try:
                from tqdm import tqdm
                return tqdm(documents, total=total)
            except:
                return documents

    @classmethod
    def create_sample_document(self, document_id: str=None, include_chunks=False):
        """
        Create sample document.
        Args:
            Document_id: the index of the document
        """
        rand_index = np.random.randint(0, 30)
        if include_chunks:
            sample_document =  {
                'color': [random.choice(['red', 'blue', 'orange', 'green'])],
                'number': [random.choice(list(range(10)))],
                'country': random.choice(['Italy', 'Australia', 'Denmark', 'Brazil', 'France']),
                'color_vector_': np.random.rand(1, 30).tolist()[0],
                'color_2_vector_': np.random.rand(1, 30).tolist()[0],
                'size': {
                    'feet': list(range(1, 31))[rand_index],
                    'cm': (np.array(range(30)) * 30.48).tolist()[rand_index]
                },
                'chunk':[
                    {
                        'color_chunkvector_': np.random.rand(1, 30).tolist()[0],
                        'color': random.choice(['red', 'blue', 'orange', 'green'])
                    },
                    {
                        'color_2_chunkvector_': np.random.rand(1, 30).tolist()[0],
                        'color': random.choice(['red', 'blue', 'orange', 'green'])
                    }
                ]
            }
        else:
            sample_document =  {
                'color': [random.choice(['red', 'blue', 'orange', 'green'])],
                'number': [random.choice(list(range(10)))],
                'country': random.choice(['Italy', 'Australia', 'Denmark', 'Brazil', 'France']),
                'color_vector_': np.random.rand(1, 30).tolist()[0],
                'color_2_vector_': np.random.rand(1, 30).tolist()[0],
                'size': {
                    'feet': list(range(1, 31))[rand_index],
                    'cm': (np.array(range(30)) * 30.48).tolist()[rand_index]
                }
            }
        if document_id is not None:
            sample_document.update({'_id': str(document_id)})
        return sample_document

    @classmethod
    def create_sample_documents(self, num_of_documents: int, include_chunks=False):
        """
        Create sample documents.
        Args:
            num_of_documents:
                Create sample documents.
        """
        return [self.create_sample_document(i, include_chunks=include_chunks) for i in range(num_of_documents)]

    @staticmethod
    def convert_concat_list_to_html(list_input):
        string = ''
        for x in list_input:
            string += '<div class="column">' + x + '</div>'
        return string

    def render_chunk(self, row, render_func):
        concat_images = [render_func(x) for x in row]
        return UtilsMixin.convert_concat_list_to_html(concat_images)

    def render_image_chunk(self, row, image_width=120):
        render_image_chunks = partial(self.render_image_in_html, image_width=image_width)
        return self.render_chunk(row, render_image_chunks)

    def render_audio_chunk(self, row):
        return self.render_chunk(row, self.render_audio_in_html)

    def show_df(self, df: pd.DataFrame,
    image_fields: List[str]=[], audio_fields: List[str]=[], chunk_image_fields: List[str]=[],
    chunk_audio_fields: List[str]=[], image_width: int=60,
    include_vector: bool=False, return_html: bool=False):
        """
            Shows a dataframe with the images and audio included inside the dataframe.
            Args:
                df:
                    Pandas DataFrame
                image_fields:
                    List of fields with the images
                audio_fields:
                    List of fields for the audio
                nrows:
                    Number of rows to preview
                image_width:
                    The width of the images
                include_vector:
                    If True, includes the vector fields
        """
        render_image_with_width = partial(self.render_image_in_html, image_width=image_width)
        formatters = {image:render_image_with_width for image in image_fields}
        formatters.update({audio: self.render_audio_in_html for audio in audio_fields})
        formatters.update({chunk_image: self.render_image_chunk for chunk_image in chunk_image_fields})
        formatters.update({chunk_audio: self.render_audio_chunk for chunk_audio in chunk_audio_fields})
        if not include_vector:
            cols = [x for x in list(df.columns) if '_vector_' not in x]
            df = df[cols]
        try:
            if return_html:
                return df.to_html(escape=False, formatters=formatters)
            from IPython.core.display import HTML
            return HTML(df.to_html(escape=False ,formatters=formatters))
        except ImportError:
            return df

    def render_image_in_html(self, path, image_width) -> str:
        return '<img src="'+ path + f'" width="{image_width}" >'

    def render_audio_in_html(self, path) -> str:
        return f"<audio controls><source src='{x}' type='audio/{self.get_audio_format(x)}'></audio>"

    def show_styler(self, df,
    image_fields: List[str]=[], audio_fields: List[str]=[], image_width: int=60,
    return_as_html=False):
        """
            Shows a dataframe with the images and audio included inside the dataframe.
            Args:
                df:
                    Pandas DataFrame
                image_fields:
                    List of fields with the images
                audio_fields:
                    List of fields for the audio
                nrows:
                    Number of rows to preview
                image_width:
                    The width of the images
        """
        render_image_with_width = partial(self.render_image_in_html, image_width=image_width)
        formatters = {image:render_image_with_width for image in image_fields}
        formatters.update({audio: self.render_audio_in_html for audio in audio_fields})

        try:
            if return_as_html:
                return df.format(formatters).render()
            return df.format(formatters)
        except ImportError:
            return df

    def get_audio_format(self, string):
        if '.wav' in string:
            return 'wav'
        if '.mp3' in string:
            return 'mpeg'
        if '.ogg' in string:
            return 'ogg'
        warnings.warn("Unable to detect audio format. Must be either wav/mpeg/ogg.")
        return ''

    def unnest_json(self, json_data, schema):
        unnested_json = {}
        for k in schema:
            unnested_json[k] = self.get_field_across_documents(k, json_data)
        return unnested_json

    def show_json(self, json: dict, selected_fields: List[str]=None, image_fields: List[str]=[],
        audio_fields: List[str]=[], chunk_image_fields: List[str]=[], chunk_audio_fields: List[str]=[],
        nrows: int=None, image_width: int=60, include_vector=False):
        """
            Shows the JSON with the audio and images inside a dataframe for quicker analysis.
            Args:
                json:
                    Dictionary
                selected_fields:
                    List of fields to see in the dictionary
                image_fields:
                    List of fields with the images
                audio_fields:
                    List of fields for the audio
                chunk_image_fields:
                    List of image fields that should be chunked
                chunk_audio_fields:
                    List of audio fields that should be chunked
                nrows:
                    Number of rows to preview
                image_width:
                    The width of the images
                include_vector:
                    Include the vector fields when showing JSON
        """
        if selected_fields is None:
            json = self.unnest_json(self.clean_results(json), schema=image_fields + audio_fields)
        else:
            json = self.unnest_json(self.clean_results(json), schema=image_fields + audio_fields + selected_fields)
        if nrows is not None:
            json = json[:nrows]
        if selected_fields is None and len(image_fields) == 0 and len(audio_fields) == 0:
            return self.show_df(pd.DataFrame(json), image_fields=image_fields, audio_fields=audio_fields,
            chunk_image_fields=chunk_image_fields, chunk_audio_fields=chunk_audio_fields,
            image_width=image_width, include_vector=include_vector)
        return self.show_df(pd.DataFrame(json),
            image_fields=image_fields, audio_fields=audio_fields,
            chunk_image_fields=chunk_image_fields, chunk_audio_fields=chunk_audio_fields,
            image_width=image_width, include_vector=include_vector)

    def show_chunk_json(self, json: dict, selected_fields: List[str]=None, image_fields: List[str]=[],
    audio_fields: List[str]=[], nrows: int=5, image_width: int=60, include_vector=False):
        """Show results if the documents are chunked.
        For images, concatenates the chunk images into the same numpy array
        For text, puts them one after the other with smaller index.
        No Audio chunking for now.
        """
        raise NotImplementedError

def get_random_int(low=0, high=9999):
    """
        Return a random int from 0 to 9999
        Args:
            low:
                Lower boundary
            High:
                Higher boundary
    """
    return random.randint(low, high)

def decorate_functions_by_argument(function_decorator, argument):
    """
        Decorate a function with an argument
        Args:
            function_decorator:
                A decorator function in Python
            argument:
                The Python argument that has to be altered
        Returns:
            A decorator
    """
    def decorator(cls):
        for name, obj in inspect.getmembers(cls):
            if isinstance(obj, (types.MethodType, types.FunctionType)):
                if obj.__name__ == "__init__":
                    continue
                if argument in inspect.signature(obj).parameters.keys():
                    try:
                        obj = obj.__func__  # unwrap Python 2 unbound method
                    except AttributeError:
                        pass  # not needed in Python 3
                    setattr(cls, name, function_decorator(obj, getattr(cls, argument)))
        return cls
    return decorator

def get_stack_level() -> int:
    """
        Return the stack level in Python.
    """
    return len(inspect.stack(0))

def set_default_collection(func, collection_name):
    """
        Set the default collection in functions
        Args:
            Func:
                The function which contains the collection_name argument
            Collection Name:
                The name of the collection
        Returns:
            Decorated function for default collection
    """
    @wraps(func)
    def wrapper(*args, **kw):
        # Set a stack level so this only effects highest level functions
        # and not internal functions that use collection_name
        if 'collection_name' in inspect.getfullargspec(func).args \
            and getattr(args[0], 'decorator_called') is False:
                setattr(args[0], 'decorator_called', True)
                setattr(args[0], 'decorator_call_stack_level',
                get_stack_level())
                num_of_args = len(inspect.getfullargspec(func).args)
                collection_name = getattr(args[0], 'collection_name')
                if collection_name in kw.keys():
                    collection_name = kw.pop('collection_name')
                # from functools import partial
                # new_func = partial(func, collection_name=collection_name)
                # print(args)
                # res = new_func(*args, **kw)
                # try:
                res = func(args[0], collection_name=collection_name, *args[1:], **kw)
                # except TypeError:
                #     res = func(args[0], collection_name=collection_name, *args[1:], **kw)
            # res = func(args[0], collection_name, *args[1:], **kw)
            # res = func(args[0], kw['collection_name'], *args[1:])
            # except:
            #     if len(args) == len(getfullargspec(func).args):
            #         res = func(*args)
            #     else:
            #         try:
            #             res = func(*args, **kw)
            #         except TypeError:
            #             kw.pop('collection_name')
            #             res = func(*args, **kw)
        else:
            try:
                res = func(*args, **kw)
            except TypeError:
                collection_name = getattr(args[0], 'collection_name')
                res = func(args[0], collection_name, *args[1:], **kw)

        if get_stack_level() == getattr(args[0], 'decorator_call_stack_level') and \
            getattr(args[0], 'decorator_called'):
            setattr(args[0], 'decorator_called', False)
        return res
    return wrapper



import io
import base64
import requests
from typing import Dict, List
from .utils import retry, return_curl_or_response

class ViClusterClient:
    """
    Clustering
    """
    def __init__(self, username, api_key, url=None):
        self.username = username
        self.api_key = api_key
        if url:
            self.url = url
        else:
            self.url = "https://api.vctr.ai"
    
    @retry()
    def clustering_job(
        self,
        collection_name: str,
        vector_field: str,
        n_clusters: int = 0,
        refresh: bool = True,
        return_curl=False,
        **kwargs
    ):
        """
Clusters a collection by a vector field

Clusters a collection into groups using unsupervised machine learning. Clusters can then be aggregated to understand whats in them and how vectors are seperating data into different groups.
        
Args:
	vector_field:
		Vector field to perform clustering on
	n_clusters:
		Number of clusters
	refresh:
		Whether to refresh the whole collection and retrain the cluster model
	collection_name:
		Name of Collection
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "vector_field": vector_field,
            "n_clusters": n_clusters,
            "refresh": refresh,
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/collection/jobs/cluster".format(self.url),
            params=params
        )
        return return_curl_or_response(response, return_type='json', return_curl=return_curl)

    @retry()
    def cluster_aggregate(
        self,
        collection_name: str,
        aggregation_query: Dict,
        page: int = 1,
        page_size: int = 10,
        asc: bool = False,
        flatten: bool = True,
        return_curl=False,
        **kwargs
    ):
        """
Aggregate every cluster in a collection

Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.

Only can be used after a vector field has been clustered with /cluster.
    
Args:
	collection_name:
		Name of Collection
	aggregation_query:
		Aggregation query to aggregate data
	page_size:
		Size of each page of results
	page:
		Page of the results
	asc:
		Whether to sort results by ascending or descending order
    flatten:
        Whether to flatten the aggregated results into a list of dictionarys or dictionary of lists.
"""
        params = {
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "aggregation_query": aggregation_query,
                "page": page,
                "page_size": page_size,
                "asc": asc,
                "flatten" : flatten
        }
        params.update(kwargs)
        response = requests.post(
            url="{}/collection/cluster_aggregate".format(self.url),
            json=params,
        )
        return return_curl_or_response(response, return_type='json', return_curl=return_curl)

    @retry()
    def cluster_facets(self, collection_name: str, facets_fields: List = [], asc: bool = True,
    page_size: int=1000, page: int=1, date_interval: str="monthly", return_curl: bool=False):
        """
Get Facets in each cluster in a collection

Takes a high level aggregation of every field and every cluster in a collection. This helps you interpret each cluster and what is in them.

Only can be used after a vector field has been clustered with /cluster.
    
Args:
	facets_fields:
		Fields to include in the facets, if [] then all
	page_size:
		Size of facet page
	page:
		Page of the results
	asc:
		Whether to sort results by ascending or descending order
	date_interval:
		Interval for date facets
	collection_name:
		Name of Collection
    date_interval:
        Defaults "monthly". Interval for date facets
"""
        params = {
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "facets_fields": facets_fields,
                "page_size": page_size,
                "page": page,
                "asc": asc,
                "date_interval": date_interval
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/collection/cluster_facets".format(self.url),
            params=params
        )
        return return_curl_or_response(response, return_type='json', return_curl=return_curl)

    @retry()
    def cluster_centroids(self, collection_name: str, vector_field: str, return_curl: bool=False, **kwargs):
        """
Returns the cluster centers of a collection by a vector field

Only can be used after a vector field has been clustered with /cluster.
    
Args:
	vector_field:
		Clustered vector field
	collection_name:
		Name of Collection
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "vector_field": vector_field,
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/collection/cluster_centroids".format(self.url),
            params=params
        )
        return return_curl_or_response(response, return_type='json', return_curl=return_curl)

    @retry()
    def cluster_centroid_documents(
        self,
        collection_name: str,
        vector_field: str,
        metric: str = "cosine",
        include_vector: bool = True,
        return_curl: bool=False,
        **kwargs
    ):
        """
Returns the document closest to each cluster center of a collection

Only can be used after a vector field has been clustered with /cluster.
    
Args:
	vector_field:
		Clustered vector field
	metric:
		Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
	include_vector:
		Include vectors in the search results
	collection_name:
		Name of Collection
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "vector_field": vector_field,
            "metric": metric,
            "include_vector": include_vector,
        }
        response = requests.get(
            url="{}/collection/cluster_centroid_documents".format(self.url),
            params=params
        )
        return return_curl_or_response(response, return_type='json', return_curl=return_curl)

    @return_curl_or_response('json')
    @retry()
    def advanced_clustering_job(
        self,
        collection_name: str,
        vector_field: str,
        alias: str='default',
        n_clusters: int = 0,
        n_init: int = 5,
        n_iter: int = 10,
        refresh: bool = True,
        return_curl: bool=False
    ):
        """
Clusters a collection by a vector field

Clusters a collection into groups using unsupervised machine learning. Clusters can then be aggregated to understand whats in them and how vectors are seperating data into different groups. 
Advanced cluster allows for more parameters to tune and alias to name each differently trained clusters.
        
Args:
	vector_field:
		Vector field to perform clustering on
	alias:
		Alias is used to name a cluster
	n_clusters:
		Number of clusters
	n_iter:
		Number of iterations in each run
	n_init:
		Number of runs to run with different centroid seeds
	refresh:
		Whether to refresh the whole collection and retrain the cluster model
	collection_name:
		Name of Collection
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "vector_field": vector_field,
            "alias": alias,
            "n_clusters": n_clusters,
            "n_init": n_init,
            "n_iter": n_iter,
            "refresh": refresh,
        }
        return requests.get(
            url="{}/collection/jobs/advanced_cluster".format(self.url),
            params=params
        )

    @return_curl_or_response('json')
    @retry()
    def advanced_cluster_aggregate(
        self,
        collection_name: str,
        aggregation_query: Dict,
        vector_field: str,
        alias: str = "default",
        page: int = 1,
        page_size: int = 10,
        asc: bool = False,
        filters: list = [],
        flatten:bool=True,
        return_curl=False,
        **kwargs
    ):
        """
Aggregate every cluster in a collection

Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.

Only can be used after a vector field has been clustered with /advanced_cluster.
    
Args:
	collection_name:
		Name of Collection
	aggregation_query:
		Aggregation query to aggregate data
	page_size:
		Size of each page of results
	page:
		Page of the results
	asc:
		Whether to sort results by ascending or descending order
	vector_field:
		Clustered vector field
	alias:
		Alias of a cluster
    flatten:
        Whether to flatten the aggregated results into a list of dictionarys or dictionary of lists.
"""
        params = {
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "vector_field" : vector_field,
                "aggregation_query": aggregation_query,
                "page": page,
                "page_size": page_size,
                "asc": asc,
                "alias": alias,
                "filters" : filters,
                "flatten" : flatten
        }
        return requests.post(
            url="{}/collection/advanced_cluster_aggregate".format(self.url),
            json=params
        )

    @return_curl_or_response('json')
    @retry()
    def advanced_cluster_facets(
        self,
        collection_name: str,
        vector_field: str,
        alias: str = "default",
        facets_fields: List = [],
        asc: bool = True,
        page_size: int=1000,
        return_curl: bool=False,
        **kwargs
    ):
        """
Get Facets in each cluster in a collection

Takes a high level aggregation of every field and every cluster in a collection. This helps you interpret each cluster and what is in them.

Only can be used after a vector field has been clustered with /advanced_cluster.
    
Args:
	vector_field:
		Clustered vector field
	alias:
		Alias is used to name a cluster
	facets_fields:
		Fields to include in the facets, if [] then all
	page_size:
		Size of facet page
	page:
		Page of the results
	asc:
		Whether to sort results by ascending or descending order
	date_interval:
		Interval for date facets
	collection_name:
		Name of Collection
"""
        params = {
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": collection_name,
                "vector_field": vector_field,
                "facets_fields": facets_fields,
                "alias": alias,
        }
        params.update(kwargs)
        return requests.get(
            url="{}/collection/advanced_cluster_facets".format(self.url),
            params=params,
        )

    @return_curl_or_response('json')
    @retry()
    def advanced_cluster_centroids(
        self, collection_name: str, vector_field: str, alias: str = "default", **kwargs
    ):
        """
Returns the cluster centers of a collection by a vector field

Only can be used after a vector field has been clustered with /advanced_cluster.
    
Args:
	vector_field:
		Clustered vector field
	alias:
		Alias is used to name a cluster
	collection_name:
		Name of Collection
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "vector_field": vector_field,
            "alias": alias,
        }
        params.update(kwargs)
        return requests.get(
            url="{}/collection/advanced_cluster_centroids".format(self.url),
            params=params
        )

    @return_curl_or_response('json')
    @retry()
    def advanced_cluster_centroid_documents(
        self,
        collection_name: str,
        vector_field: str,
        alias: str = "default",
        metric: str = "cosine",
        include_vector: bool = True,
        return_curl: bool=False,
        **kwargs
    ):
        """
Returns the document closest to each cluster center of a collection

Only can be used after a vector field has been clustered with /advanced_cluster.
    
Args:
	vector_field:
		Clustered vector field
	alias:
		Alias is used to name a cluster
	metric:
		Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
	include_vector:
		Include vectors in the search results
	collection_name:
		Name of Collection
"""
        params={
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "vector_field": vector_field,
            "metric": metric,
            "include_vector": include_vector,
            "alias": alias,
        }
        params.update(kwargs)
        return requests.get(
            url="{}/collection/advanced_cluster_centroid_documents".format(self.url),
            params=params
        )



import io
import base64
import requests
from .base import ViDeployedModel


class ViAudio2Vec(ViDeployedModel):
    def encode(self, audio):
        return requests.get(
            url="{}/collection/encode_audio".format(self.url),
            params={
                "username": self.username,
                "api_key": self.api_key,
                "collection_name": self.collection_name,
                "audio_url": audio,
            },
        ).json()

    @property
    def __name__(self):
        if self._name is None:
            return "vectorai_audio"
        return self._name

    @__name__.setter
    def __name__(self, value):
        self._name = value


class ViAudioArray2Vec(ViDeployedModel):
    def __init__(
        self,
        username,
        api_key,
        url=None,
        collection_name="base",
        vector_operation: str = "mean",
    ):
        self.username = username
        self.api_key = api_key
        if url:
            self.url = url
        else:
            self.url = "https://api.vctr.ai"
        self.collection_name = collection_name

    def encode(self, audios):
        return _vector_operation(
            [
                requests.get(
                    url="{}/collection/encode_audio".format(self.url),
                    params={
                        "username": self.username,
                        "api_key": self.api_key,
                        "collection_name": self.collection_name,
                        "audio_url": audio,
                    },
                ).json()
                for audio in audios
            ],
            vector_operation=self.vector_operation,
        )

    @property
    def __name__(self):
        if self._name is None:
            return "vectorai_audio_array"
        return self._name

    @__name__.setter
    def __name__(self, value):
        self._name = value



"""Scoring Functions to compare for analytics.
"""
from typing import Dict, Any, List
from ..read import ViReadClient
from numpy import inner
from numpy.linalg import norm

class ViScore(ViReadClient):
    @staticmethod
    def calculate_cosine_similarity(a, b):
        return inner(a, b) / (norm(a) * norm(b))
    
    def get_cosine_similarity_scores(
        self,
        other_documents: List[Dict[str, Any]],
        anchor_document: Dict[str, Any],
        vector_field: str,
    ) -> List[float]:
        """
        Compare scores based on cosine similarity

        Args:
            other_documents:
                List of documents (Python Dictionaries)
            anchor_document:
                Document to compare all the other documents with.
            vector_field:
                The field in the documents to compare
        
        Example:
            >>> documents = [{...}]
            >>> ViClient.get_cosine_similarity_scores(documents[1:10], documents[0])
        """
        similarity_scores = []
        for i, doc in enumerate(other_documents):
            similarity_score = self.calculate_cosine_similarity(
                self.get_field(vector_field, doc),
                self.get_field(vector_field, anchor_document)
            )
            similarity_scores.append(similarity_score)
        return similarity_scores



from .api import _ViAPIClient
from .cluster import *
from .text import *
from .audio import *
from .image import *
from .dimensionality_reduction import *

class ViAPIClient(_ViAPIClient, ViDimensionalityReductionClient, ViTextClient, ViAudioClient, ViImageClient, ViClusterClient):
    def __init__(self, username, api_key, url="https://api.vctr.ai"):
        self.username = username
        self.api_key = api_key
        self.url = url



"""Vi Dimensionality Reduction
"""
import numpy as np
import pandas as pd
from typing import List, Dict, Any
from ..write import ViWriteClient


class ViDimensionalityReductionBase(ViWriteClient):
    """
    Base class for dimensionality reduction.
    
    Example:
        >>> class IVISDimensionalityReduction(ViDimensionalityReductionBase):
        >>>     def __init__(self, batch_size=120, k=15, embedding_dims=2):
        >>>         self.scaler = MinMaxScaler()
        >>>         self.model = Ivis(embedding_dims=embedding_dims, k=k, batch_size=batch_size)
        >>>     def fit_transform(self, documents, field_vector):
        >>>           if isinstance(documents[0], dict):
        >>>             if 'documents' in documents[0].keys():
        >>>                vectors = [self.get_field(field_vector, document) for document in documents['documents']]
        >>>            else:
        >>>                vectors = [self.get_field(field_vector, document) for document in documents]
        >>>            
        >>>            X = np.stack(vectors)
        >>>            X_scaled = self.scaler.fit_transform(X)
        >>>            return self.model.fit_transform(X_scaled)
        >>>    
        >>>        def transform(self, X):
        >>>            if isinstance(documents[0], dict):
        >>>                if 'documents' in documents[0].keys():
        >>>                    vectors = [document[field_vector] for document in documents['documents']]
        >>>                else:
        >>>                    vectors = [document[field_vector] for document in documents]
        >>>            X_scaled = self.scaler.transform(vectors)
        >>>            return self.model.transform(X_scaled)
    """
    
    def fit_transform(self, documents: list, vector_field: str):
        """
        Fit and transform

        Args:
            documents:
                A list of lists to flatten to make into 1 list.
            vector_field:
                The vector field to fit and transform
        Returns:
            A dimensionality reduced array:
                A numpy array with less dimensions than the original.

        Example:

        >>> class IVISDimensionalityReduction(ViDimensionalityReductionBase):
        >>>     def __init__(self, batch_size=120, k=15, embedding_dims=2):
        >>>         self.scaler = MinMaxScaler()
        >>>         self.model = Ivis(embedding_dims=embedding_dims, k=k, batch_size=batch_size)
        >>>     def fit_transform(self, documents, field_vector):
        >>>           if isinstance(documents[0], dict):
        >>>             if 'documents' in documents[0].keys():
        >>>                vectors = [self.get_field(field_vector, document) for document in documents['documents']]
        >>>            else:
        >>>                vectors = [self.get_field(field_vector, document) for document in documents]
        >>>            
        >>>            X = np.stack(vectors)
        >>>            X_scaled = self.scaler.fit_transform(X)
        >>>            return self.model.fit_transform(X_scaled)
        >>>    
        >>>        def transform(self, X):
        >>>            if isinstance(documents[0], dict):
        >>>                if 'documents' in documents[0].keys():
        >>>                    vectors = [document[field_vector] for document in documents['documents']]
        >>>                else:
        >>>                    vectors = [document[field_vector] for document in documents]
        >>>            X_scaled = self.scaler.transform(vectors)
        >>>            return self.model.transform(X_scaled)
        """
        pass
    
    def transform(self, documents: List[Dict[str, Any]]):
        """
        Transform documents

        Args:
            documents:
                A list of lists to flatten to make into 1 list.
        
        Returns:
            transformed array:
                A transformed numpy array

        Example:
        
        >>> class IVISDimensionalityReduction(ViDimensionalityReductionBase):
        >>>     def __init__(self, batch_size=120, k=15, embedding_dims=2):
        >>>         self.scaler = MinMaxScaler()
        >>>         self.model = Ivis(embedding_dims=embedding_dims, k=k, batch_size=batch_size)
        >>>     def fit_transform(self, documents, field_vector):
        >>>           if isinstance(documents[0], dict):
        >>>             if 'documents' in documents[0].keys():
        >>>                vectors = [self.get_field(field_vector, document) for document in documents['documents']]
        >>>            else:
        >>>                vectors = [self.get_field(field_vector, document) for document in documents]
        >>>            
        >>>            X = np.stack(vectors)
        >>>            X_scaled = self.scaler.fit_transform(X)
        >>>            return self.model.fit_transform(X_scaled)
        >>>    
        >>>        def transform(self, X):
        >>>            if isinstance(documents[0], dict):
        >>>                if 'documents' in documents[0].keys():
        >>>                    vectors = [document[field_vector] for document in documents['documents']]
        >>>                else:
        >>>                    vectors = [document[field_vector] for document in documents]
        >>>            X_scaled = self.scaler.transform(vectors)
        >>>            return self.model.transform(X_scaled)
        """
        pass



import requests
import random
from typing import Dict, List
from .audio import ViAudioClient
from .image import ViImageClient
from .text import ViTextClient
from .cluster import ViClusterClient
from .search import ViSearchClient
from .dimensionality_reduction import ViDimensionalityReductionClient
from .array_dict_vectorizer import ViArrayDictClient
from .utils import retry, return_curl_or_response

class ViReadAPIClient(
    ViSearchClient,
    ViTextClient,
    ViAudioClient,
    ViImageClient,
    ViClusterClient,
    ViDimensionalityReductionClient,
    ViArrayDictClient
):
    """
    Read Operations
    """
    def __init__(self, username, api_key, url=None):
        self.username = username
        self.api_key = api_key
        if url:
            self.url = url
        else:
            self.url = "https://api.vctr.ai"
            
    @retry()
    def _list_collections(self, return_curl: bool=False, **kwargs):
        """
Lists all the collections in a project

Retrieve a list all the created collections.
    
Args:
	username:
		Username
	api_key:
		Api Key, you can request it from request_api_key
"""
        params = {
            "username": self.username, 
            "api_key": self.api_key
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/project/list_collections".format(self.url),
            params=params,
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def collection_stats(self, collection_name: str, return_curl: bool=False, **kwargs):
        """
Retrieves stats about a collection

Stats include: size, searches, number of documents, etc.
    
Args:
	collection_name:
		Name of Collection
"""

        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/project/collection_stats".format(self.url),
            params=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def collection_schema(self, collection_name: str, return_curl: bool=False, **kwargs):
        """
Retrieves the schema of a collection

The schema of a collection can include types of: text, numeric, date, bool, etc.
    
Args:
	collection_name:
		Name of Collection
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name
        }
        params.update(kwargs)

        response = requests.get(
            url="{}/project/collection_schema".format(self.url),
            params=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def id(self, collection_name: str, document_id: str, include_vector: bool = True, return_curl: bool=False, **kwargs):
        """
Look up a document by its id
    
Args:
	document_id:
		ID of a document
	include_vector:
		Include vectors in the search results
	collection_name:
		Name of Collection
"""

        params={
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "document_id": document_id,
            "include_vector": include_vector,
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/collection/id".format(self.url),
            params=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def bulk_id(self, collection_name: str, document_ids: List[str], return_curl: bool=False, **kwargs):
        """
Look up multiple document by their ids
    
Args:
	document_ids:
		IDs of documents
	include_vector:
		Include vectors in the search results
	collection_name:
		Name of Collection
"""

        params={
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "document_ids": document_ids,
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/collection/bulk_id".format(self.url),
            params=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def retrieve_documents(
        self,
        collection_name: str,
        page_size: int = 20,
        cursor: str = None,
        sort: List = [],
        asc: bool = True,
        include_vector: bool = True,
        include_fields: List = [],
        return_curl: bool=False,
        **kwargs
    ):
        """
Retrieve some documents

Cursor is provided to retrieve even more documents. Loop through it to retrieve all documents in the database.
    
Args:
	include_fields:
		Fields to include in the document, if empty list [] then all is returned
	cursor:
		Cursor to paginate the document retrieval
	page_size:
		Size of each page of results
	sort:
		Fields to sort the documents by
	asc:
		Whether to sort results by ascending or descending order
	include_vector:
		Include vectors in the search results
	collection_name:
		Name of Collection
"""
        q_params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "include_fields": include_fields,
            "page_size": page_size,
            "cursor": cursor,
            "asc": asc,
            "include_vector": include_vector,
        }
        if sort:
            q_params["sort"] = sort
        q_params.update(kwargs)
        response = requests.get(
            url="{}/collection/retrieve_documents".format(self.url), params=q_params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def random_documents(
        self,
        collection_name: str,
        page_size: int = 20,
        seed: int = None,
        include_vector: bool = True,
        include_fields: list = [],
        return_curl: bool=False,
        **kwargs
    ):
        """
Retrieve some documents randomly

Mainly for testing purposes.
    
Args:
	seed:
		Random Seed for retrieving random documents.
	page_size:
		Size of each page of results
	include_vector:
		Include vectors in the search results
	collection_name:
		Name of Collection
    return_curl:
        Return CURL statement
"""
        if seed is None:
            seed = random.randint(0, 9999)
        q_params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "page_size": page_size,
            "seed": seed,
            "include_vector": include_vector,
            "include_fields": include_fields
        }
        q_params.update(kwargs)

        response = requests.get(
            url="{}/collection/random_documents".format(self.url), params=q_params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def id_lookup_joined(self, join_query: dict, doc_id:str, return_curl: bool=False, **kwargs):
        """
Look up a document by its id with joins
    
Args:
	join_query:
		.
	doc_id:
		ID of a Document
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "join_query": join_query,
            "doc_id" : doc_id,
        }
        params.update(kwargs)
        response = requests.post(
            url="{}/collection/id_lookup_joined".format(self.url),
            json=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)
        
    @retry()
    def aggregate(
        self,
        collection_name: str,
        aggregation_query: Dict,
        page:int=1,
        page_size:int=10,
        asc:bool=False,
        flatten:bool=True,
        return_curl: bool=False,
        **kwargs
    ):
        """
Aggregate a collection

Aggregation/Groupby of a collection using an aggregation query. 
The aggregation query is a json body that follows the schema of:: 

    {
        "groupby" : [
            {"name": <nickname/alias>, "field": <field in the collection>, "agg": "category"},
            {"name": <another_nickname/alias>, "field": <another groupby field in the collection>, "agg": "category"}
        ], 
        "metrics" : [
            {"name": <nickname/alias>, "field": <numeric field in the collection>, "agg": "avg"}
        ]
    }

- "groupby" is the fields you want to split the data into. These are the available groupby types:

    - category" : groupby a field that is a category
- "metrics" is the fields you want to metrics you want to calculate in each of those. These are the available metric types: every aggregation includes a frequency metric:

    - average", "max", "min", "sum", "cardinality"
    
Args:
	collection_name:
		Name of Collection
	aggregation_query:
		Aggregation query to aggregate data
	page_size:
		Size of each page of results
	page:
		Page of the results
	asc:
		Whether to sort results by ascending or descending order
    flatten:
        Whether to flatten the aggregated results into a list of dictionarys or dictionary of lists.
    return_curl:
        Return the CURL statement
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "aggregation_query": aggregation_query,
            "page": page,
            "page_size": page_size,
            "asc": asc,
            "flatten" : flatten
        }
        params.update(kwargs)
        response = requests.post(
            url="{}/collection/aggregate".format(self.url),
            json=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def facets(
        self,
        collection_name: str,
        fields: List[str] = [],
        page: int = 1,
        page_size: int = 20,
        asc: bool = False,
        return_curl: bool=False,
        **kwargs
    ):
        """
Retrieve the facets of a collection

Takes a high level aggregation of every field in a collection. This is used in advance search to help create the filter bar for search.
    
Args:
	facets_fields:
		Fields to include in the facets, if [] then all
	date_interval:
		Interval for date facets
	page_size:
		Size of facet page
	page:
		Page of the results
	asc:
		Whether to sort results by ascending or descending order
	collection_name:
		Name of Collection
"""

        params={
            "username": self.username,
            "api_key": self.api_key,
            "facets_fields" : fields,
            "collection_name": collection_name,
            "page": page,
            "page_size": page_size,
            "asc": asc,
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/collection/facets".format(self.url),
            params=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def filters(
        self,
        collection_name: str,
        filters: List,
        page=1,
        page_size=10,
        include_vector: bool = False,
        return_curl: bool=False,
        **kwargs
    ):
        """
Filters a collection

Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched.

The filters query is a json body that follows the schema of::

    [
        {'field' : <field to filter>, 'filter_type' : <type of filter>, "condition":"==", "condition_value":"america"},
        {'field' : <field to filter>, 'filter_type' : <type of filter>, "condition":">=", "condition_value":90},
    ]

These are the available filter_type types::

    1. "contains": for filtering documents that contains a string.
            {'field' : 'category', 'filter_type' : 'contains', "condition":"==", "condition_value": "bluetoo"]}
    2. "exact_match"/"category": for filtering documents that matches a string or list of strings exactly.
            {'field' : 'category', 'filter_type' : 'categories', "condition":"==", "condition_value": "tv"]}
    3. "categories": for filtering documents that contains any of a category from a list of categories.
            {'field' : 'category', 'filter_type' : 'categories', "condition":"==", "condition_value": ["tv", "smart", "bluetooth_compatible"]}
    4. "exists": for filtering documents that contains a field.
            {'field' : 'purchased', 'filter_type' : 'exists', "condition":">=", "condition_value":" "}
    5. "date": for filtering date by date range.
            {'field' : 'insert_date_', 'filter_type' : 'date', "condition":">=", "condition_value":"2020-01-01"}
    6. "numeric": for filtering by numeric range. 
            {'field' : 'price', 'filter_type' : 'date', "condition":">=", "condition_value":90}

These are the available conditions:
 
    "==", "!=", ">=", ">", "<", "<="
    
Args:
	collection_name:
		Name of Collection
	filters:
		Query for filtering the search results
	page:
		Page of the results
	page_size:
		Size of each page of results
	asc:
		Whether to sort results by ascending or descending order
	include_vector:
		Include vectors in the search results
    return_curl:
        Returns Curl statement for debugging
"""
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "filters": filters,
            "page": page,
            "page_size": page_size,
            "include_vector": include_vector,
        }
        params.update(kwargs)
        response = requests.post(
            url="{}/collection/filters".format(self.url),
            json=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def job_status(self, collection_name: str, job_id: str, job_name: str, return_curl: bool=False, **kwargs):
        """
Get status of a job. Whether its starting, running, failed or finished.
    
Args:
	job_id:
		.
	job_name:
		.
	collection_name:
		Name of Collection
"""

        params={
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "job_id": job_id,
            "job_name": job_name,
        }
        response = requests.get(
            url="{}/collection/jobs/job_status".format(self.url),
            params=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def list_jobs(self, collection_name: str, return_curl: bool=False, **kwargs):
        """
Get history of jobs

List and get a history of all the jobs and its job_id, parameters, start time, etc.
    
Args:
	collection_name:
		Name of Collection
"""

        params={
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
        }
        params.update(kwargs)
        response = requests.get(
            url="{}/collection/jobs/list_jobs".format(self.url),
            params=params
        )
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def bulk_missing_id(self, collection_name: str, document_ids: List[str], return_curl: bool=False, **kwargs):
        """
            Return IDs that are not in a collection.
        """
        params = {
            "username" : self.username,
            "api_key" : self.api_key,
            "collection_name": collection_name,
            "document_ids" : document_ids
        }
        params.update(kwargs)
        response = requests.post('{}/collection/bulk_missing_id'.format(self.url), 
        json=params)
        return return_curl_or_response(response, 'json', return_curl=return_curl)
    
    @retry()
    def random_documents_with_filters(self, collection_name: str,
    seed: int=None, include_fields: List[str]=[], page_size: int=20,
    include_vector: bool=False, filters: List[Dict]=[], return_curl: bool=False, **kwargs):
        """
            Random documents with filters.
        """

        if seed is None:
            seed = random.randint(0, 9999)

        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name,
            "seed": seed,
            "include_fields": include_fields,
            "page_size": page_size,
            "include_vector": include_vector,
            "filters": filters
        }

        response = requests.post('{}/collection/random_documents_with_filters'.format(self.url), json=params)
        return return_curl_or_response(response, 'json', return_curl=return_curl)

    @retry()
    def vector_health(self, collection_name: str, return_curl: bool=False, **kwargs):
        """Return vector health of a collection
        """
        params = {
            "username": self.username,
            "api_key": self.api_key,
            "collection_name": collection_name
        }
        params.update(kwargs)
        response = requests.get("{}/collection/vector_health".format(self.url), params=params)
        return return_curl_or_response(response, 'json', return_curl=return_curl)
        
